#ifndef HRK_CONNECTION_H
#define HRK_CONNECTION_H

/*!
  \file
  \brief 接続のインターフェース

  \author Satofumi Kamimura

  $Id$

  \todo timeout を double 型にする。
*/

#include <cstddef>


namespace hrk
{
    /*!
      \brief 接続のインターフェース

      アクセス用メソッドを定義する。ただし open() は扱うデバイスによって引数の数が変わる可能性があるため、この実装を行うインターフェースが提供ことにする。
    */
    class Connection
    {
    public:
        enum {
            Timeout_infinity = -1, //!< read() で無限に受信を待つ場合に利用する
        };


        virtual ~Connection(void)
        {
        }

        /*!
          \brief 状態を示すメッセージを返す

          \return メッセージ文字列
        */
        virtual const char* what(void) const = 0;

        /*!
          \brief 受信

          \retval true 接続中
          \retval false 接続していない
        */
        virtual bool is_open(void) const = 0;

        /*!
          \brief 接続を閉じる

          接続されてないときに呼び出されたときは、何もしない。
        */
        virtual void close(void) = 0;


        /*!
          \brief ボーレートの変更

          \param[in] baudrate ボーレート [bps]

          \retval true 成功
          \retval false エラー
        */
        virtual bool change_baudrate(long baudrate) = 0;

        /*!
          \brief データの送信

          \param[in] data 送信するデータ
          \param[in] data_size 送信するデータの byte 数

          \retval >=0 送信した byte 数
          \retval <0 エラー ID
        */
        virtual int write(const char* data, size_t data_size) = 0;

        /*!
          \brief データの受信

          \param[in] data 受信データ用のバッファ
          \param[in] max_data_size 受信できるデータの最大 byte 数
          \param[in] timeout 受信を待つ時間 [msec]。無限に待つ場合には Timeout_infinity を指定する。

          \retval >=0 送信した byte 数
          \retval <0 エラー ID
        */
        virtual int read(char* data, size_t max_data_size, int timeout) = 0;


        /*!
          \brief １文字だけ受信バッファに書き戻す
        */
        virtual void ungetc(int ch) = 0;
    };
}

#endif
